#!/usr/bin/env python2

################################################################################
# Copyright (C) 2017 NICE s.r.l.
# All Rights Reserved
#
# This software is the confidential and proprietary information
# of NICE s.r.l. ("Confidential Information").
# You shall not disclose such Confidential Information
# and shall use it only in accordance with the terms of
# the license agreement you entered into with NICE.
################################################################################

"""
DCV Simple External Authenticator
Usage::
    dcvsimpleextauth.py -h
Send a POST request:
    curl -k https://localhost:<port> -d sessionId=<session-id> -d authenticationToken=<token>
"""

import sys
try:
    assert sys.version_info[0] == 2
except:
    sys.exit(format("ERROR: Python version %s not supported!" % sys.version_info[0]))

from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from SocketServer import ThreadingMixIn
from ConfigParser import SafeConfigParser

import ConfigParser
import os.path
import ssl
import cgi
import sys
import select
import hashlib
import datetime

#  Default values
DEFAULT_LOG_DIR = "/var/log/dcv"
DEFAULT_PORT = 8444


class Handler(BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        self.server.log_file.write("%s - - [%s] %s\n" %
                         (self.address_string(),
                          datetime.datetime.utcnow(),
                          format%args))
        self.server.log_file.flush()

    def _set_headers_200(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/xml')
        self.end_headers()

    def _return_auth_ok(self, username):
        self._set_headers_200()
        self.wfile.write('<auth result="yes"><username>' + username + '</username></auth>')

    def _return_auth_ko(self):
        self._set_headers_200()
        self.wfile.write('<auth result="no"><message>I do not know who you are!</message></auth>')

    def do_GET(self):
        self.send_error(405, "Unable to process GET request, use POST instead")

    def do_HEAD(self):
        self._set_headers_200()

    def do_POST(self):
        if '127.0.0.1' in self.client_address or 'localhost' in self.client_address:
            # Parse the form data posted
            form = cgi.FieldStorage(
                fp=self.rfile,
                headers=self.headers,
                environ={'REQUEST_METHOD': 'POST',
                         'CONTENT_TYPE': self.headers['Content-Type'],
                         })

            if form.length < 2048:
                session_id_req = ''
                auth_token_req = ''

                if "sessionId" in form:
                    session_id_req = form.getvalue("sessionId")
                if "authenticationToken" in form:
                    auth_token_req = form.getvalue("authenticationToken")

                if session_id_req and auth_token_req:
                    is_auth = self._check_auth(session_id_req=session_id_req, token_req=auth_token_req)
                    if is_auth:
                        self._return_auth_ok(username=is_auth)
                    else:
                        self._return_auth_ko()
                else:
                    self.log_error("ERROR: POST Request for session with Id (%s) does not contain all needed parameters." % session_id_req)
                    self._return_auth_ko()
            else:
                self.log_error("ERROR: POST Request too long")
                self._return_auth_ko()
        else:
            self.log_error("ERROR: POST Request not coming from localhost")
            self._return_auth_ko()

    def _check_auth(self, session_id_req, token_req):
        session_file_path = self.server.auth_dir + '/' + session_id_req
        if os.path.isfile(session_file_path):
            try:
                auth_parser = SafeConfigParser()
                auth_parser.read(session_file_path)

                if auth_parser.has_section("auth"):
                    username_value = auth_parser.get('auth', hash_token(token_req, session_id_req))

                    if username_value:
                        self.log_message("Accepted request authentication for session with Id (%s)" % session_id_req)
                        return username_value
                    else:
                        self.log_message("Invalid request authentication for session with Id (%s)" % session_id_req)
                else:
                    self.log_error("ERROR: Unable to find [auth] section in authentication file for session with Id (%s)" % session_id_req)
            except ConfigParser.NoOptionError as ex:
                self.log_error("ERROR: Unable to find authentication token for session with Id (%s)" % session_id_req)
            except Exception as ex:
                self.log_error("ERROR: Unable to parse authentication file for session with Id (%s), with exception (%s)" % (session_id_req, ex.__class__.__name__))
        else:
            self.log_error("ERROR: Unable to find authentication file for session with Id (%s)" % session_id_req)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""


def run_server(port, certificate, auth_dir, log_dir, server_class=ThreadedHTTPServer, handler_class=Handler):
    server_address = ('localhost', port)
    httpd = server_class(server_address, handler_class)
    # FIXME add log rotation
    log_file_path = log_dir + '/dcvsimpleextauth.log'
    log_file = open(log_file_path, 'w')
    os.chmod(log_file_path, 0o600)
    httpd.log_file = log_file
    httpd.auth_dir = auth_dir
    protocol = "HTTP"
    if certificate:
        httpd.socket = ssl.wrap_socket(httpd.socket, certfile=certificate, server_side=True)
        protocol = "HTTPS"
    print('Starting DCV external authenticator %s server on port %s, use <Ctrl-C> to stop' % (protocol, port))
    httpd.serve_forever()


def run_remove_auth(session, auth_dir):
    session_file_path = auth_dir + '/' + session

    if os.path.isfile(session_file_path):
        os.remove(session_file_path)
    else:
        sys.exit(format("ERROR: Unable to delete auth file for session (%s). Does not exist" % session))


def run_add_user(user, session, auth_dir, append):
    session_file_path = auth_dir + '/' + session

    timeout = 2.0
    if not select.select([sys.stdin,],[],[],timeout)[0]:
        sys.exit("ERROR: Unable to read token from stdin")

    plain_token = sys.stdin.read().strip()

    if append and os.path.isfile(session_file_path):
        auth_file = open(session_file_path, 'a')
        auth_file.write(hash_token(plain_token, session) + " = " + user + "\n")
        auth_file.close()
    else:
        auth_file = open(session_file_path, 'w')
        auth_file.write("[auth]\n")
        auth_file.write(hash_token(plain_token, session) + " = " + user + "\n")
        auth_file.close()
        os.chmod(session_file_path, 0o640)


def hash_token(token, session):
    return sha256_checksum(token + session)


def sha256_checksum(string):
    sha256 = hashlib.sha256()
    sha256.update(string)
    return sha256.hexdigest()


if __name__ == "__main__":
    import argparse

    # Instantiate the parser
    parser = argparse.ArgumentParser(description='DCV simple external authenticator server')

    # FIXME better help
    parser.add_argument('command', choices=['start-server', 'add-user', 'remove-auth'], help="The command to execute: start-server (launch the simple external authenticator web server), add-user (add authentication token for a user on the specified session) or remove-auth (remove any authentication token from the specified session)")

    # command start-server parameters
    parser.add_argument('--certificate', help="Valid for start-server command. For HTTPS, the certificate file to be used by the authenticator server")
    parser.add_argument('--port', default=DEFAULT_PORT, type=int, help="Valid for start-server command. The port to listen for authentication request")
    parser.add_argument('--log-dir', default=DEFAULT_LOG_DIR, help="Valid for start-server command. The folder where log files are stored")

    # command add-user parameters
    parser.add_argument('--user', help="Valid for add-user command. The user to add to the session authentication file")
    parser.add_argument('--session', help="Valid for add-user and remove-auth command. The session id where authentication changes apply")
    parser.add_argument('--append', action='store_true', default=False, help="Valid for add-user command. Append the specified user to the session authentication file")

    # common parameters
    parser.add_argument('--auth-dir', required=True, help="Valid for start-server, add-user and remove-auth command. The folder where session authentication files are stored")

    args = parser.parse_args()

    is_server = False
    is_add_user = False
    is_remove_auth = False

    if args.command == 'start-server':
        is_server = True
    elif args.command == 'add-user':
        is_add_user = True
    elif args.command == 'remove-auth':
        is_remove_auth = True
    else:
        sys.exit("ERROR: Command not supported!")

    if is_server:
        run_server(port=args.port, certificate=args.certificate, auth_dir=args.auth_dir, log_dir=args.log_dir)
    elif is_add_user:
        run_add_user(user=args.user, session=args.session, auth_dir=args.auth_dir, append=args.append)
    elif is_remove_auth:
        run_remove_auth(session=args.session, auth_dir=args.auth_dir)
    else:
        sys.exit("ERROR: This should never happen!")

# ex:ts=4:et:
